/*
 * Copyright (C) 2010 InfinitiesSoft Corporation. 
 * http://www.infinitiessoft.com
 * 
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
*/
package org.jboss.seam.contexts;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.jboss.seam.Component;
import org.jboss.seam.ScopeType;
import org.jboss.seam.core.Manager;
import org.jboss.seam.log.Log;
import org.jboss.seam.log.Logging;

/**
 * A enhanced lifecycle controller for running seam in multiple war
 * 
 * @author cola.orange@gmail.com
 * @author Dennis.Chen
 */
public class LifecycleEx {

    private static String NULL_APP_NAME = "LifecycleEx.NULL.APPNAME";
    private static Map<String, Object> defaultApplication = null;
    private static Map<String, Map<String, Object>> applications = new HashMap<String, Map<String, Object>>();
    private static List<Map<String, Object>> applicationList = new LinkedList<Map<String, Object>>();
    private static final Log log = Logging.getLog(LifecycleEx.class);

    static public void registerApplication(String name, Map<String, Object> application) {

        name = name == null ? NULL_APP_NAME : name;

        if (applications.containsKey(name)) {
            log.warn("trying to register application '#0' again", name);
        }

        applications.put(name, application);
        applicationList.add(application);

        defaultApplication = application;
    }

    static public void unregisterApplication(String name) {
        name = name == null ? NULL_APP_NAME : name;

        Map<String, Object> app = applications.remove(name);

        if (app != null) {
            applicationList.remove(app);
            if (app == defaultApplication) {
                if (applicationList.size() > 0) {
                    // last one
                    defaultApplication = applicationList.get(applicationList.size() - 1);
                } else {
                    defaultApplication = null;
                }
            }
        }
    }

    static Map<String, Object> getApplication(String name) {
        return getApplication(name, false);
    }

    /**
     * get Application context by name, if no application context found, then
     * try to get default application context (the last registered). if there is
     * no registered one, then it get application context form original
     * Lifecycle.getApplication()
     * 
     * 
     * @param name
     *            name of application
     * @param as_possible
     *            true: if get application by name not found, then return
     *            default application or form Lifecycle.getApplication()
     * @return RuntimeException if no application found.
     */
    static Map<String, Object> getApplication(String name, boolean as_possible) {
        name = name == null ? NULL_APP_NAME : name;

        Map<String, Object> app = applications.get(name);
        if (app == null && as_possible) {
            log.debug("name #0 not found, use default application", name);
            app = defaultApplication;
        }
        if (app == null && as_possible) {
            log.debug("default application not found, use seam Lifecycle.getApplication", name);
            app = Lifecycle.getApplication();
        }
        if (app == null) {
            throw new RuntimeException("application " + name + " not found ");
        }
        return app;
    }

    public static void setupApplication(String name) {
        Contexts.applicationContext.set(new ApplicationContext(getApplication(name)));
    }

    public static void cleanupApplication() {
        Contexts.applicationContext.set(null);
    }

    static void clearThreadlocals() {
        Contexts.eventContext.set(null);
        Contexts.pageContext.set(null);
        Contexts.sessionContext.set(null);
        Contexts.conversationContext.set(null);
        Contexts.businessProcessContext.set(null);
        Contexts.applicationContext.set(null);
    }

    public static void beginCall(String appname) {
        beginCall(appname, false);
    }

    public static void beginCall(String appname, boolean as_possible) {
        log.debug(">>> Begin call #0",appname);
        Contexts.applicationContext.set(new ApplicationContext(getApplication(appname, as_possible)));
        Contexts.eventContext.set(new BasicContext(ScopeType.EVENT));
        // TODO should we do session scope initialization??
        Contexts.sessionContext.set(new BasicContext(ScopeType.SESSION));
        Contexts.conversationContext.set(new BasicContext(ScopeType.CONVERSATION));
        Contexts.businessProcessContext.set(new BusinessProcessContext());
    }

    public static void endCall() {
        try {
            Contexts.destroy(Contexts.getSessionContext());
            Contexts.flushAndDestroyContexts();
            if (Manager.instance().isLongRunningConversation()) {
                throw new IllegalStateException("Do not start long-running conversations in direct calls to EJBs");
            }
        } finally {
            clearThreadlocals();
            log.debug("<<< End call");
        }
    }

    public static void wrapEventContext(Map<String, Object> context) {
        if (!Contexts.isEventContextActive()) {
            throw new IllegalStateException("Event context is not active");
        }
        Context evtCtx = Contexts.getEventContext();
        Contexts.eventContext.set(new ContextWrapper(ScopeType.EVENT, context, evtCtx));
    }
    
    public static void unwrapEventContext(){
        Context evtCtx = Contexts.getEventContext();
        if(evtCtx instanceof ContextWrapper){
            Contexts.eventContext.set(((ContextWrapper)evtCtx).context);
        }
    }

    public static class ContextWrapper implements Context,Serializable {
        private static final long serialVersionUID = 1L;
        
        Context context;
        ScopeType scope;
        Map<String, Object> map;

        protected ContextWrapper(ScopeType scope, Map<String, Object> map, Context context) {
            this.scope = scope;
            this.map = map;
            this.context = context;
        }

        public void flush() {
            context.flush();
        }

        public Object get(String name) {
            Object obj = map.get(name);
            return obj != null ? obj : context.get(name);
        }

        public ScopeType getType() {
            return scope;
        }

        public Object get(Class clazz) {
            return get(Component.getComponentName(clazz));
        }

        public String[] getNames() {
            ArrayList<String> keys = new ArrayList<String>();
            for (String key : map.keySet())
            {
               keys.add(key);
            }
            for(String key:context.getNames()){
                keys.add(key);
            }
            return keys.toArray(new String[keys.size()]);
            
        }

        public boolean isSet(String name) {
            return map.containsKey(name) || context.isSet(name);
        }

        public void remove(String name) {
            if(map.containsKey(name)){
                map.remove(name);
            }
            if(context.isSet(name)){
                context.remove(name);
            }
        }

        public void set(String name, Object value) {
            if(map.containsKey(name)){
                map.put(name,value);
            }else{
                context.set(name,value);
            }
        }
        
        public String toString()
        {
           return "ContextWrapper(" + scope + ")";
        }

    }
}
