/*
 * Copyright (C) 2010 InfinitiesSoft Corporation. 
 * http://www.infinitiessoft.com
 * 
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
*/
package org.jboss.seam.contexts;

import java.util.Map;

import javax.servlet.ServletContext;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.jboss.seam.core.ConversationPropagation;
import org.jboss.seam.core.Init;
import org.jboss.seam.core.Manager;
import org.jboss.seam.log.Log;
import org.jboss.seam.log.Logging;
import org.jboss.seam.servlet.ServletApplicationMap;
import org.jboss.seam.servlet.ServletRequestMap;
import org.jboss.seam.servlet.ServletRequestSessionMap;
import org.jboss.seam.transaction.Transaction;
import org.jboss.seam.web.ServletContexts;
import org.jboss.seam.web.Session;

/**
 * A enhanced servlet lifecycle controller for running seam in multiple war
 * 
 * @author cola.orange@gmail.com
 * @author Dennis.Chen
 */
public class ServletLifecycleEx {

    private static final Log log = Logging.getLog(ServletLifecycleEx.class);

    static public void registerApplication(String name, ServletContext context) {
        log.info("Register application context name '#0'", name);
        Map<String, Object> application = new ServletApplicationMap(context);
        LifecycleEx.registerApplication(name, application);

    }

    static public void unregisterApplication(String name) {
        log.info("Unregister application context name '#0'", name);
        LifecycleEx.unregisterApplication(name);

    }

    public static void beginRequest(String name, HttpServletRequest request) {
        log.debug("begin web request '#0'", name);
        
        Contexts.eventContext.set(new EventContext(new ServletRequestMap(request)));
        Contexts.sessionContext.set(new SessionContext(new ServletRequestSessionMap(request)));
        Contexts.applicationContext.set(new ApplicationContext(LifecycleEx.getApplication(name)));
        Contexts.conversationContext.set(null); // in case endRequest() was
                                                // never called
        
        ServletContexts.instance().setRequest(request);
    }

    public static void endRequest(HttpServletRequest request) {
        log.debug("After request, destroying contexts");
        try {
            Session session = Session.getInstance();
            boolean sessionInvalid = session != null && session.isInvalid();

            Contexts.flushAndDestroyContexts();

            if (sessionInvalid) {
                LifecycleEx.clearThreadlocals();
                request.getSession().invalidate();
                // actual session context will be destroyed from the listener
            }
        } finally {
            LifecycleEx.clearThreadlocals();
            log.debug("<<< End web request");
        }
    }
    
    
    public static void handleTransactionsAfterRequest(boolean rollback) {
        if (Init.instance().isTransactionManagementEnabled()) {
            try {
                if (Transaction.instance().isActive() && !rollback) {
                    try {
                        log.debug("committing transaction ");
                        Transaction.instance().commit();

                    } catch (IllegalStateException e) {
                        log.warn("TX commit failed with illegal state exception. This may be "
                                + "because the tx timed out and was rolled back in the background.", e);
                    }
                } else if (Transaction.instance().isRolledBackOrMarkedRollback() || rollback) {
                    log.debug("rolling back transaction ");
                    Transaction.instance().rollback();
                }

            } catch (Exception e) {
                throw new IllegalStateException("Could not commit transaction", e);
            }
        }

    }

    public static void handleTransactionsBeforeRequest() {
        if (Init.instance().isTransactionManagementEnabled()) {
            try {
                if (!Transaction.instance().isActiveOrMarkedRollback()) {
                    log.debug("beginning transaction ");
                    Transaction.instance().begin();
                }
            } catch (Exception e) {
                throw new IllegalStateException("Could not start transaction", e);
            }
        }
    }
    
    /**
     * to restore a conversation, if the given conversation is not exist, then create a temporary conversation
     * @param conversationId
     * @param request
     */
    public static void restoreConversation(String conversationId,HttpServletRequest request) {
        ConversationPropagation.instance().setConversationId(conversationId);
        Manager mgr = Manager.instance();
    //  if conversation not found , it will restore a temp conversation
        boolean conversationFound = mgr.restoreConversation();
        
        ServletLifecycleEx.resumeConversation(request);
        log.debug("conversation id '#0', found '#1'", conversationId, mgr.getCurrentConversationId(),
                conversationFound);
    }
    

    private static void resumeConversation(HttpServletRequest request) {
        ServerConversationContext conversationContext = new ServerConversationContext(new ServletRequestSessionMap(
                request));
        Contexts.conversationContext.set(conversationContext);
        Contexts.businessProcessContext.set(new BusinessProcessContext());
        conversationContext.unflush();
    }
    
    public static void handleConversationAfterRequest(HttpServletRequest request){
        Manager.instance().endRequest(new ServletRequestSessionMap(request));
    }

}
